#include "ts_calibrate.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define XORMODE	0x80000000
#define XRES 240
#define YRES 320

struct ts_sample
{
    int		x;
    int		y;
    unsigned int	pressure;
};

/*
xfb[5]，yfb[5]存放预先设定的5个LCD坐标值
x[5]，y[5]存放从触摸芯片读回来的触摸坐标值
*/
typedef struct
{
    int x[5], xfb[5];
    int y[5], yfb[5];
    int a[7];
} calibration;

calibration cal;

/*******************************************************************************
 * 函数名(Function):	perform_calibration
 * 参数(Param):				cal
 * 返回值(Return):		是否校验成功
 *										0 失败
 *										1 成功
 * 描述(Description):	LCD画点函数
 *******************************************************************************/
int perform_calibration(calibration *cal)
{
    int j;
    float n, x, y, x2, y2, xy, z, zx, zy;
    float det, a, b, c, e, f, i;
    float scaling = 65536.0;

    // Get sums for matrix
    n = x = y = x2 = y2 = xy = 0;
    for(j = 0; j < 5; j++)
    {
        n += 1.0;
        x += (float)cal->x[j];
        y += (float)cal->y[j];
        x2 += (float)(cal->x[j] * cal->x[j]);
        y2 += (float)(cal->y[j] * cal->y[j]);
        xy += (float)(cal->x[j] * cal->y[j]);
    }

    // Get determinant of matrix -- check if determinant is too small
    det = n * (x2 * y2 - xy * xy) + x * (xy * y - x * y2) + y * (x * xy - y * x2);
    if(det < 0.1 && det > -0.1)
    {
        printf("ts_calibrate: determinant is too small -- %f\n", det);
        return 0;
    }

    // Get elements of inverse matrix
    a = (x2 * y2 - xy * xy) / det;
    b = (xy * y - x * y2) / det;
    c = (x * xy - y * x2) / det;
    e = (n * y2 - y * y) / det;
    f = (x * y - n * xy) / det;
    i = (n * x2 - x * x) / det;

    // Get sums for x calibration
    z = zx = zy = 0;
    for(j = 0; j < 5; j++)
    {
        z += (float)cal->xfb[j];
        zx += (float)(cal->xfb[j] * cal->x[j]);
        zy += (float)(cal->xfb[j] * cal->y[j]);
    }

    // Now multiply out to get the calibration for framebuffer x coord
    cal->a[0] = (int)((a * z + b * zx + c * zy) * (scaling));
    cal->a[1] = (int)((b * z + e * zx + f * zy) * (scaling));
    cal->a[2] = (int)((c * z + f * zx + i * zy) * (scaling));

    printf("%f %f %f\n", (a * z + b * zx + c * zy),
           (b * z + e * zx + f * zy),
           (c * z + f * zx + i * zy));

    // Get sums for y calibration
    z = zx = zy = 0;
    for(j = 0; j < 5; j++)
    {
        z += (float)cal->yfb[j];
        zx += (float)(cal->yfb[j] * cal->x[j]);
        zy += (float)(cal->yfb[j] * cal->y[j]);
    }

    // Now multiply out to get the calibration for framebuffer y coord
    cal->a[3] = (int)((a * z + b * zx + c * zy) * (scaling));
    cal->a[4] = (int)((b * z + e * zx + f * zy) * (scaling));
    cal->a[5] = (int)((c * z + f * zx + i * zy) * (scaling));

    printf("%f %f %f\n", (a * z + b * zx + c * zy),
           (b * z + e * zx + f * zy),
           (c * z + f * zx + i * zy));

    // If we got here, we're OK, so assign scaling to a[6] and return
    cal->a[6] = (int)scaling;
    return 1;

}

/*******************************************************************************
 * 函数名(Function):	put_cross
 * 参数(Param):				x 坐标X
 *										y 坐标Y
 *										colidx 颜色
 * 返回值(Return):		无
 * 描述(Description):	输出对准点
 *******************************************************************************/
void put_cross(int x, int y, unsigned colidx)
{
    //	int i,j;
    //				for(i = -3; i < 4; i++)
    //				{
    //			  	for(j = -3; j < 4; j++)
    //						LCD_PutPixel(LCD_PANEL_UPPER,x+i,y+j,RED);
    //				}


    LCD_DrawLine (x - 10, y, x - 2, y, 0x07e0);
    LCD_DrawLine (x + 2, y, x + 10, y, 0x07e0);
    LCD_DrawLine (x, y - 10, x, y - 2, 0x07e0);
    LCD_DrawLine (x, y + 2, x, y + 10, 0x07e0);

    /*
    	LCD_DrawLine (x - 6, y - 9, x - 9, y - 9,0x07e0);
    	LCD_DrawLine (x - 9, y - 8, x - 9, y - 6,0x07e0);
    	LCD_DrawLine (x - 9, y + 6, x - 9, y + 9,0x07e0);
    	LCD_DrawLine (x - 8, y + 9, x - 6, y + 9,0x07e0);
    	LCD_DrawLine (x + 6, y + 9, x + 9, y + 9,0x07e0);
    	LCD_DrawLine (x + 9, y + 8, x + 9, y + 6,0x07e0);
    	LCD_DrawLine (x + 9, y - 6, x + 9, y - 9,0x07e0);
    	LCD_DrawLine (x + 8, y - 9, x + 6, y - 9,0x07e0);
    */
}

#define MAX_SAMPLES 20
struct ts_sample samp[MAX_SAMPLES];

/*******************************************************************************
 * 函数名(Function):	getxy
 * 参数(Param):				x 坐标X
 *										y 坐标Y
 * 返回值(Return):		是否采样成功
 *										0 失败
 *										1	成功
 * 描述(Description):	采样X，Y坐标数据
 *										读取20组触摸坐标值，进行排序取中值平均，并进行阈值检查
 *******************************************************************************/
int getxy(int *x, int *y)
{
    int i, j, tmp;
    int index, middle;
    int timeout;

    timeout = 0;
    /* 检测是否有触摸按下 */
    while(GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8) && (timeout < 20))
    {
        timeout++;
    }
    /* 若超时则返回 */
    if(timeout >= 20)
        return 0;

    //	*x = GUI_TOUCH_X_MeasureX();
    //	*y = GUI_TOUCH_X_MeasureY();
    //	return 1;

    index = 0;
    /* 采样MAX_SAMPLES组数据直到触摸松开，并保存到数组中 */
    while((index < MAX_SAMPLES) && (!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8)))
    {
        index++;

        samp[index].x = SPI_TOUCH_Read_X();
        samp[index].y = SPI_TOUCH_Read_Y();
    }

    printf("Took %d samples...\n", index);
    if(index < MAX_SAMPLES)
        return 0;

    middle = index / 2;
    if (x)
    {
        /* 对采样到的数据进行冒泡法排序 */
        for(i = 0; i < MAX_SAMPLES; i++)
        {
            for(j = 0; j < MAX_SAMPLES - i - 1; j++)
            {
                if(samp[j + 1].x < samp[j].x)
                {
                    tmp = samp[j + 1].x;
                    samp[j + 1].x = samp[j].x;
                    samp[j].x = tmp;
                }
            }
        }
        if(samp[middle].x - samp[middle - 1].x > 5)
            return 0;
        if (index & 1)	//若为奇数个点，则取中间3个点平均
            *x = (samp[middle - 1].x + samp[middle].x + samp[middle + 1].x) / 3;
        else						//否则取中间4个点平均
            *x = (samp[middle - 1].x + samp[middle].x + samp[middle - 2].x + samp[middle + 1].x) / 4;
    }
    if (y)
    {
        /* 对采样到的数据进行冒泡法排序  */
        for(i = 0; i < MAX_SAMPLES; i++)
        {
            for(j = 0; j < MAX_SAMPLES - i - 1; j++)
            {
                if(samp[j + 1].y < samp[j].y)
                {
                    tmp = samp[j + 1].y;
                    samp[j + 1].y = samp[j].y;
                    samp[j].y = tmp;
                }
            }
        }
        if(samp[middle].y - samp[middle - 1].y > 5)
            return 0;
        if (index & 1)	//若为奇数个点，则取中间3个点平均
            *y = (samp[middle - 1].y + samp[middle].y + samp[middle + 1].y) / 3;
        else						//否则取中间4个点平均
            *y = (samp[middle - 1].y + samp[middle].y + samp[middle - 2].y + samp[middle + 1].y) / 4;
    }
    return 1;

}

/*******************************************************************************
 * 函数名(Function):	get_sample
 * 参数(Param):				cal
 *										index 采样点序号
 *										x 坐标X
 *										y 坐标Y
 *										name
 * 返回值(Return):		无
 * 描述(Description):	采样数据并保存到cal中
 *******************************************************************************/
static void get_sample (calibration *cal,
                        int index, int x, int y, char *name)
{

    put_cross(x, y, 2 | XORMODE);
    /* 调用getxy将采样到的触摸坐标值存放在cal->x[]，cal->y[]数组里 */
    while(!getxy (&cal->x [index], &cal->y [index]));
    put_cross(x, y, 2 | XORMODE);

    /* 将预先设定的LCD坐标值存放在xfb[]，yfb[]数组里 */
    cal->xfb [index] = x;
    cal->yfb [index] = y;

    printf("%s : LCD X = %4d Y = %4d\n", name, cal->xfb [index], cal->yfb [index]);
    printf("%s : X = %4d Y = %4d\n", name, cal->x [index], cal->y [index]);
}

/*******************************************************************************
 * 函数名(Function):	Phys2Log
 * 参数(Param):				PhysX	触摸坐标X
 *										PhysY	触摸坐标Y
 *										LogX LCD坐标X
 *										LogY LCD坐标Y
 * 返回值(Return):		无
 * 描述(Description):	物理坐标转换为逻辑坐标（触摸坐标转换为LCD坐标）
 *******************************************************************************/
void Phys2Log(int PhysX, int PhysY, int *LogX, int *LogY)
{

    printf("before Calibration x = %d y=%d\n", PhysX, PhysY);
    *LogX = (int)(( cal.a[0] + cal.a[1] * PhysX + cal.a[2] * PhysY ) / cal.a[6]);
    *LogY =	(int)(( cal.a[3] + cal.a[4] * PhysX + cal.a[5] * PhysY ) / cal.a[6]);
    printf("after Calibration x = %d y=%d\n", *LogX, *LogY);

}

/*******************************************************************************
 * 函数名(Function):	ts_calibrate
 * 参数(Param):				无
 * 返回值(Return):		是否校准成功
 *										-1 失败
 *										0	成功
 * 描述(Description):	执行触摸校准程序
 *******************************************************************************/
int ts_calibrate(void)
{
    char cal_buffer[256];
    int i;

    /* 清除LCD屏幕 */
    LCD_Clear(0xffff);
    get_sample (&cal, 0, 50,        50,        "Top left");
    while(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8));

    /* 清除LCD屏幕 */
    LCD_Clear(0xffff);
    get_sample (&cal, 1, XRES - 50, 50,        "Top right");
    while(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8));

    /* 清除LCD屏幕 */
    LCD_Clear(0xffff);
    get_sample (&cal, 2, XRES - 50, YRES - 50, "Bot right");
    while(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8));

    /* 清除LCD屏幕 */
    LCD_Clear(0xffff);
    get_sample (&cal, 3, 50,        YRES - 50, "Bot left");
    while(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8));

    /* Clear the LCD */
    LCD_Clear(0xffff);
    get_sample (&cal, 4, XRES / 2,  YRES / 2,  "Center");
    while(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8));

    /* 清除LCD屏幕 */
    LCD_Clear(0xffff);

    /* 把参数代入公式，执行校准程序，得到校准参数 */
    if (perform_calibration (&cal))
    {
        printf ("Calibration constants: ");
        for (i = 0; i < 7; i++) printf("%d ", cal.a [i]);
        printf("\n");

        sprintf (cal_buffer, "%d %d %d %d %d %d %d", cal.a[0], cal.a[1], cal.a[2], cal.a[3], cal.a[4], cal.a[5], cal.a[6]);
        printf ("%d %d %d %d %d %d %d", cal.a[1], cal.a[2], cal.a[0], cal.a[4], cal.a[5], cal.a[3], cal.a[6]);

        i = 0;
    }
    else
    {
        printf("Calibration failed.\n");
        i = -1;
    }

    return i;
}


static uint16_t x_tolerance = 0;
static uint16_t y_tolerance = 0;
/*******************************************************************************
 * 函数名(Function):	ts_test
 * 参数(Param):				无
 * 返回值(Return):		无
 * 描述(Description):	触摸画板测试
 *******************************************************************************/
void ts_test(void)
{
    int x_cur, y_cur;
    int x, y;
    int i, j;
    while(1)
    {
        /* 判断触摸按下 */
        if(!GPIO_ReadInputDataBit(GPIOE, GPIO_Pin_8))
        {
            /* 采样物理坐标 */
            if(getxy (&x_cur, &y_cur))
            {
                /* 物理坐标转换为逻辑坐标 */
                Phys2Log(x_cur, y_cur, &x, &y);
                if(x - x_tolerance > 15 || x_tolerance - x > 15)
                {
                    if(y - y_tolerance > 15 || y_tolerance - y > 15)
                    {
                        x_tolerance = x;
                        /* 输出小正方型点 */
                        for(i = -2; i < 3; i++)
                        {
                            for(j = -2; j < 3; j++)
                                LCD_DrawPoint(x + i, y + j, 0xf800);
                        }

                        printf("Mouse x = %d,y = %d \n", x * 1920 / 240, y * 1080 / 320);
                    }
                }
            }
            else
            {
                printf("Mouse x = %d,y = %d \n", y_tolerance * 1920 / 240, y_tolerance * 1080 / 320);
                /* 输出小正方型点 */
                for(i = -2; i < 3; i++)
                {
                    for(j = -2; j < 3; j++)
                        LCD_DrawPoint(x + i, y + j, 0xf800);
                }
            }
        }
    }

}


/************************************* End *************************************/
